#!/usr/bin/python3
# -*- coding: utf-8 -*-

from utils_fonction import create_matrix, create_matrix_with_affine_penalty, INIT_GAP, EXTEND_GAP
from score import Score
from sequence import Sequence

class GlobalAlign:

    def __init__(self, sequenceA, sequenceB, substitutionMatrix):

        self.sequenceA = sequenceA
        self.sequenceB = sequenceB
        self.alignSequenceA = ""
        self.alignSequenceB = ""
        self.substitutionMatrix = Score.load(substitutionMatrix)
        self.solutions = []
        self.lineSize = len(self.sequenceA)
        self.columnSize = len(self.sequenceB)
        self.S = create_matrix_with_affine_penalty(self.lineSize, self.columnSize)
        self.V = create_matrix(self.lineSize, self.columnSize)
        self.W = create_matrix(self.lineSize, self.columnSize)

        for idcol in range(self.columnSize+1):
            self.V[0][idcol] = -float("inf")
        for idraw in range(self.lineSize+1):
            self.W[idraw][0] = -float("inf")

        for idraw in range(self.lineSize+1):
            for idcol in range(self.columnSize+1):
                if idraw  > 0 and idcol == 0:
                    self.V[idraw][idcol] = max(self.S[idraw-1][idcol] - INIT_GAP,
                                               self.V[idraw-1][idcol] - EXTEND_GAP)
                if idraw == 0 and idcol > 0:
                    self.W[idraw][idcol] = max(self.S[idraw][idcol - 1] - INIT_GAP,
                                               self.W[idraw][idcol - 1] - EXTEND_GAP)
                elif idraw > 0 and idcol > 0:
                    self.V[idraw][idcol] = max(self.S[idraw - 1][idcol] - INIT_GAP,
                                               self.V[idraw - 1][idcol] - EXTEND_GAP)
                    self.W[idraw][idcol] = max(self.S[idraw][idcol-1] - INIT_GAP, self.W[idraw][idcol-1] - EXTEND_GAP)
                    self.S[idraw][idcol] = max(
                        self.S[idraw-1][idcol-1]+self.substitutionMatrix.getScore(self.sequenceA[idraw-1], self.sequenceB[idcol-1]),
                        self.V[idraw][idcol],
                        self.W[idraw][idcol])


        self.backtrackMatrixTrace = self.createBacktrackMatrixTrace()
        self.globalAlign(self.lineSize, self.columnSize)
        self.printSolution()

    def printSolution(self):
        print("Résultat obtenue pour un alignement global entre la séquence A: {} et la séquence B: {} \n".format(
                                                                                        self.sequenceA, self.sequenceB))

        print("Score: {} \nAlignement globaux obtenus :\n".format(self.S[-1][-1]))
        for alignSequences in self.solutions:
            print("{}\n{}\n".format(alignSequences[0],alignSequences[1]))



    def globalAlign(self, idraw, idcol):

        if idraw == 0 and idcol == 0:
            #Cas d'arrêt
            self.solutions.append((self.alignSequenceA[::-1], self.alignSequenceB[::-1]))

        else:
            #Cas récursif/backtracking
            for way in self.backtrackMatrixTrace[idraw][idcol]:
                temp_idraw, temp_idcol = idraw, idcol
                if way == "T":
                    self.alignSequenceA += self.sequenceA[idraw-1]
                    self.alignSequenceB += "-"
                    idraw-=1
                elif way == "L":
                    self.alignSequenceA += "-"
                    self.alignSequenceB += self.sequenceB[idcol-1]
                    idcol-= 1
                else:
                    self.alignSequenceA += self.sequenceA[idraw-1]
                    self.alignSequenceB += self.sequenceB[idcol-1]
                    idraw-=1
                    idcol-=1
                self.globalAlign(idraw, idcol)
                self.alignSequenceA = self.alignSequenceA[:-1]
                self.alignSequenceB = self.alignSequenceB[:-1]
                idraw, idcol = temp_idraw, temp_idcol

    def createBacktrackMatrixTrace(self):
        matrix = [["" for i in range(self.columnSize+1)] for j in range(self.lineSize+1)]
        matrix[0][0] = "0"
        for idraw in range(self.lineSize+1):
            for idcol in range(self.columnSize+1):
                if idraw == 0 and idcol > 0:
                    matrix[idraw][idcol] = "L"
                if idraw > 0 and idcol == 0:
                    matrix[idraw][idcol] = "T"
                if idraw > 0 and idcol > 0:
                    if self.S[idraw][idcol] == self.V[idraw][idcol]:
                        matrix[idraw][idcol] += "T"
                    if self.S[idraw][idcol] == self.W[idraw][idcol]:
                        matrix[idraw][idcol] += "L"
                    if self.S[idraw][idcol] == self.S[idraw-1][idcol-1]+self.substitutionMatrix.getScore(self.sequenceA[idraw-1], self.sequenceB[idcol-1]):
                        matrix[idraw][idcol] += "D"
        return matrix


if __name__ == '__main__':
    sequences = Sequence.load("WW-sequence.fasta")
    #testAlign = GlobalAlign("THISLINE", "ISALIGNED", substitutionMatrix="blosum62.txt")
    #testAlign = GlobalAlign(sequences[0].getSequence(), sequences[4].getSequence(), substitutionMatrix="blosum62.txt")
    testAlign = GlobalAlign("GGVTTT", "MGGETFA", substitutionMatrix="blosum62.txt")
