from utils_fonction import create_matrix, create_matrix_with_affine_penalty, INIT_GAP, EXTEND_GAP
from score import Score
from sequence import Sequence
from copy import deepcopy
class LocalAlign:
    def __init__(self, sequenceA, sequenceB, substitutionMatrix, numberSeqToFind=1):

        self.sequenceA = sequenceA
        self.sequenceB = sequenceB
        self.lineSize = len(self.sequenceA)
        self.columnSize = len(self.sequenceB)
        self.substitutionMatrix = Score.load(substitutionMatrix)
        self.solutions = []
        self.allMaximunPos = [] #Mémoire des maximum déjà trouvé
        self.allVisitedPos = set() #Mémoire de passage bactracking
        self.S = create_matrix(self.lineSize, self.columnSize)
        self.V = create_matrix(self.lineSize, self.columnSize)
        self.W = create_matrix(self.lineSize, self.columnSize)

        for idcol in range(self.columnSize+1):
            self.V[0][idcol] = -float("inf")
        for idraw in range(self.lineSize+1):
            self.W[idraw][0] = -float("inf")

        self.fillMatrix()

        self.backtrackMatrixTrace = self.createBacktrackMatrixTrace()
        for k in range(numberSeqToFind):
            self.alignSequenceA = ""
            self.alignSequenceB = ""
            maxPos = self.findMaxPos()
            if maxPos == (0,0):
                #Matrice S contient que des 0
                # -1 = valeur sentinelle pour dire plus possible de trouver des séquences.
                self.solutions.append(-1)
                break
            else:
                self.allMaximunPos.append(maxPos)
                visitedPos = set()
                self.localAlign(maxPos, visitedPos)
                self.allVisitedPos.update(visitedPos)
                self.recreateMatrix(visitedPos)
                self.backtrackMatrixTrace = self.createBacktrackMatrixTrace()

        self.printSolution()

    def findMaxPos(self):
        maximum = 0
        maxPos = (0,0)
        for idraw in range(1, self.lineSize+1):
            for idcol in range(1, self.columnSize+1):
                if (idraw, idcol) not in self.allMaximunPos and (self.S[idraw][idcol] > maximum):
                    maxPos = (idraw, idcol)
                    maximum = self.S[idraw][idcol]
        return maxPos


    def createBacktrackMatrixTrace(self):
        matrix = [["" for i in range(self.columnSize + 1)] for j in range(self.lineSize + 1)]
        for idraw in range(self.lineSize + 1):
            for idcol in range(self.columnSize + 1):
                if self.S[idraw][idcol] != 0:
                    if idraw == 0 and idcol > 0:
                        matrix[idraw][idcol] = "L"
                    if idraw > 0 and idcol == 0:
                        matrix[idraw][idcol] = "T"
                    if idraw > 0 and idcol > 0:

                        if self.S[idraw][idcol] == self.S[idraw - 1][idcol - 1] + self.substitutionMatrix.getScore(self.sequenceA[idraw - 1], self.sequenceB[idcol - 1]):
                            matrix[idraw][idcol] += "D"

                        if self.S[idraw][idcol] == self.W[idraw][idcol]:
                            matrix[idraw][idcol] += "L"

                        if self.S[idraw][idcol] == self.V[idraw][idcol]:
                            matrix[idraw][idcol] += "T"

                else:
                    matrix[idraw][idcol] = "0"
        return matrix

    def localAlign(self, maxPos, visitedPos):

        idraw, idcol = maxPos[0], maxPos[1]
        while self.S[idraw][idcol] != 0:
            visitedPos.add((idraw, idcol))

            if self.backtrackMatrixTrace[idraw][idcol] == "T":
                self.alignSequenceA += self.sequenceA[idraw-1]
                self.alignSequenceB += "-"
                idraw-=1
            elif self.backtrackMatrixTrace[idraw][idcol] == "L":
                self.alignSequenceA += "-"
                self.alignSequenceB += self.sequenceB[idcol-1]
                idcol-= 1
            else:
                self.alignSequenceA += self.sequenceA[idraw-1]
                self.alignSequenceB += self.sequenceB[idcol-1]
                idraw-=1
                idcol-=1

        self.solutions.append((self.alignSequenceA[::-1], self.alignSequenceB[::-1], self.S[maxPos[0]][maxPos[1]]))

    def fillMatrix(self):
        for idraw in range(self.lineSize+1):
            for idcol in range(self.columnSize+1):
                tuplePos = (idraw, idcol)
                if tuplePos not in self.allVisitedPos:
                    if idraw > 0 and idcol == 0:
                        self.V[idraw][idcol] = max(self.S[idraw - 1][idcol] - INIT_GAP,
                                                   self.V[idraw - 1][idcol] - EXTEND_GAP)
                    if idraw == 0 and idcol > 0:
                        self.W[idraw][idcol] = max(self.S[idraw][idcol - 1] - INIT_GAP,
                                                   self.W[idraw][idcol - 1] - EXTEND_GAP)
                    elif idraw > 0 and idcol > 0:
                        self.V[idraw][idcol] = max(self.S[idraw - 1][idcol] - INIT_GAP,
                                                   self.V[idraw - 1][idcol] - EXTEND_GAP)
                        self.W[idraw][idcol] = max(self.S[idraw][idcol - 1] - INIT_GAP,
                                                   self.W[idraw][idcol - 1] - EXTEND_GAP)
                        self.S[idraw][idcol] = max(
                            self.S[idraw - 1][idcol - 1] + self.substitutionMatrix.getScore(self.sequenceA[idraw - 1],
                                                                                            self.sequenceB[idcol - 1]),
                            self.V[idraw][idcol],
                            self.W[idraw][idcol])
                    if self.S[idraw][idcol] < 0:
                        self.S[idraw][idcol] = 0


    def recreateMatrix(self, visitedPos):

        self.clearPath(visitedPos)
        self.fillMatrix()

    def clearPath(self, visitedPos):
        for tuplePos in visitedPos:
            self.S[tuplePos[0]][tuplePos[1]] = 0
            self.V[tuplePos[0]][tuplePos[1]] = 0
            self.W[tuplePos[0]][tuplePos[1]] = 0


    def printSolution(self):
        print("Résultat obtenue pour un alignement local entre la séquence A: {} et la séquence B: {} \n".format(
            self.sequenceA, self.sequenceB))

        print("\nAlignement locaux obtenus :\n")
        for i in range (len(self.solutions)):
            if self.solutions[i] != -1:
                print("{}\n{}\nScore:{}\n".format(self.solutions[i][0], self.solutions[i][1], self.solutions[i][2]))
            else:
                print("Impossible de trouver plus de séquences alignés")



if __name__ == '__main__':
    sequences = Sequence.load("protein-sequences.fasta")
    #testAlign = LocalAlign("THISLINE", "ISALIGNED", substitutionMatrix="blosum62.txt")
    #testAlign = LocalAlign("WRITERS","VINTNER", substitutionMatrix="blosum62.txt", numberSeqToFind=10)
    testAlign = LocalAlign(sequences[0].getSequence(), sequences[1].getSequence(), substitutionMatrix="blosum62.txt", numberSeqToFind=2)
