#!/usr/bin/python3
# -*- coding: utf-8 -*-
import re

from score import Score
from sequence import Sequence


class Align:
    def __init__(self, sequenceA, sequenceB, substitutionMatrixName, I, E ,local=False):

        self.sequenceA = sequenceA
        self.sequenceB = sequenceB
        self.lineSize = len(sequenceA)
        self.columnSize = len(sequenceB)
        self.I = I
        self.E = E
        self.substitutionMatrixName = substitutionMatrixName
        self.local = local
        self.alignSequenceA = ""
        self.alignSequenceB = ""
        self.substitutionMatrix = Score.loadData(substitutionMatrixName)
        self.S = self.initMatrix()
        self.V = self.initMatrix()
        self.W = self.initMatrix()

        for idcol in range(self.columnSize + 1):
            self.V[0][idcol] = -float("inf")
        for idraw in range(self.lineSize + 1):
            self.W[idraw][0] = -float("inf")

        self.allMaximunPos = []  # Mémoire des maximum déjà trouvé utile que pour LocalALign
        self.allVisitedPos = set()  # Mémoire de passage bactracking utile que pour LocalALign
        self.solutions = []

    def initMatrix(self):
        return [[0 for i in range(self.columnSize + 1)] for j in range(self.lineSize + 1)]

    def createBacktrackMatrixTrace(self):
        matrix = [["0" for i in range(self.columnSize + 1)] for j in range(self.lineSize + 1)]
        for idraw in range(self.lineSize + 1):
            for idcol in range(self.columnSize + 1):
                if not self.local or self.S[idraw][idcol] != 0:
                    if idraw == 0 and idcol > 0:
                        matrix[idraw][idcol] = "L"
                    if idraw > 0 and idcol == 0:
                        matrix[idraw][idcol] = "T"
                    if idraw > 0 and idcol > 0:
                        matrix[idraw][idcol] = ""
                        if self.S[idraw][idcol] == self.S[idraw - 1][idcol - 1] + self.substitutionMatrix[self.sequenceA[idraw - 1], self.sequenceB[idcol - 1]]:
                            matrix[idraw][idcol] += "D"

                        if self.S[idraw][idcol] == self.W[idraw][idcol]:
                            matrix[idraw][idcol] += "L"

                        if self.S[idraw][idcol] == self.V[idraw][idcol]:
                            matrix[idraw][idcol] += "T"
                else:
                    matrix[idraw][idcol] = "0"
        return matrix

    def fillMatrices(self):
        for idraw in range(self.lineSize + 1):
            for idcol in range(self.columnSize + 1):
                tuplePos = (idraw, idcol)
                if tuplePos not in self.allVisitedPos:
                    if idraw > 0 and idcol == 0:
                        self.V[idraw][idcol] = max(self.S[idraw - 1][idcol] - self.I,
                                                   self.V[idraw - 1][idcol] - self.E)
                    if idraw == 0 and idcol > 0:
                        self.W[idraw][idcol] = max(self.S[idraw][idcol - 1] - self.I,
                                                   self.W[idraw][idcol - 1] - self.E)
                    elif idraw > 0 and idcol > 0:
                        self.V[idraw][idcol] = max(self.S[idraw - 1][idcol] - self.I,
                                                   self.V[idraw - 1][idcol] - self.E)
                        self.W[idraw][idcol] = max(self.S[idraw][idcol - 1] - self.I,
                                                   self.W[idraw][idcol - 1] - self.E)
                        self.S[idraw][idcol] = max(
                            self.S[idraw - 1][idcol - 1] + self.substitutionMatrix[self.sequenceA[idraw - 1], self.sequenceB[idcol - 1]],
                            self.V[idraw][idcol],
                            self.W[idraw][idcol])
                if self.local and self.S[idraw][idcol] < 0:
                    self.S[idraw][idcol] = 0

    def __repr__(self):
        """
        Représentation des alignement trouvés.
        """
        stringToPrint = ""
        if self.local:
            stringToPrint +="Résultat obtenue pour un alignement local entre la séquence A: {} \n et la séquence B: {} \n\n".format(
                self.sequenceA, self.sequenceB)
            stringToPrint += "Matrice de substitution: {}\nOpen Gap Penalty: {}\nExtend Gap Penalty: {}\n\n".format(
                re.sub(r'\.txt$', '', self.substitutionMatrixName), self.I, self.E)
            stringToPrint += "Alignements locaux obtenus :\n\n"
        else:
            stringToPrint += "Résultat obtenue pour un alignement global entre la séquence A: \n{} et la séquence B: {} \n\n".format(
                self.sequenceA, self.sequenceB)
            stringToPrint += "Alignements globaux obtenus :\n\n"

        for i in range(len(self.solutions)):
            if self.solutions[i] != -1:  # condtion utile que pour alignement local
                stringToPrint += "Score: {}\n\n".format(self.solutions[i][2])
            if self.local:
                stringToPrint += "Partie de la Séquence A alignée de {} à {}\nPartie de la Séquence B alignée de {} à {} \n\n".format(self.solutions[i][3], self.solutions[i][4], self.solutions[i][5], self.solutions[i][6])
                for j in range(0, len(self.solutions[i][0]), 60):
                    stringToPrint += "Sequence A alignée: {}\n{}{}\nSequence B alignée: {}\n\n".format(self.solutions[i][0][j:j + 60], " " * 20,
                                                                          self.compareSequence(
                                                                              self.solutions[i][0][j:j + 60],
                                                                              self.solutions[i][1][j:j + 60]),
                                                                          self.solutions[i][1][j:j + 60])
            else:
                stringToPrint += "Impossible de trouver plus de séquences alignées\n"
        return stringToPrint

    def __str__(self):
        return repr(self)

    def compareSequence(self, sequenceA, sequenceB):
        """
        Construit un string pour avoir une représentation comme sur LALIGN.
        | = match parfait entre 2 acides aminés
        : = score de substitution positif ou nul entre 2 acide aminé
        . = score de substitution négatif entre 2 acide aminé
        " " = lorsqu'il y a un gap dans une des 2 séquences.
        """
        stringRes = ""
        for i in range(len(sequenceA)):
            charA = sequenceA[i]
            charB = sequenceB[i]
            if charA == charB:
                stringRes += "|"
            elif (charA != "-" and charB != "-") and self.substitutionMatrix[charA, charB] >= 0:
                stringRes += ":"
            elif (charA != "-" and charB != "-") and self.substitutionMatrix[charA, charB] < 0:
                stringRes += "."
            else:
                stringRes += " "
        return stringRes


class GlobalAlign(Align):
    def __init__(self, sequenceA, sequenceB, substitutionMatrixName="blosum62.txt", I=12, E=2, numberSeqToFind=1):

        Align.__init__(self, sequenceA, sequenceB, substitutionMatrixName, I, E)
        self.numberSeqFind = 0
        self.numberSeqToFind = numberSeqToFind
        self.createMatrixWithAffinePenalty()
        self.fillMatrices()
        self.backtrackMatrixTrace = self.createBacktrackMatrixTrace()
        self.findAlign(self.lineSize, self.columnSize)
        if self.numberSeqFind != self.numberSeqToFind:
            # -1 = valeur sentinelle pour dire plus possible de trouver des séquences.
            self.solutions.append(-1)

    def createMatrixWithAffinePenalty(self):

        for i in range(1, self.lineSize + 1):
            self.S[i][0] = -(self.I) - ((i - 1) * self.E)
        for j in range(1, self.columnSize + 1):
            self.S[0][j] = -(self.E * (j - 1)) - self.I

    def findAlign(self, idraw, idcol):

        if idraw == 0 and idcol == 0 and self.numberSeqFind < self.numberSeqToFind:
            # Cas d'arrêt
            self.solutions.append((self.alignSequenceA[::-1], self.alignSequenceB[::-1], self.S[-1][-1]))
            self.numberSeqFind += 1

        elif self.numberSeqFind < self.numberSeqToFind:
            # Cas récursif/backtracking
            for way in self.backtrackMatrixTrace[idraw][idcol]:
                temp_idraw, temp_idcol = idraw, idcol
                if way == "T":
                    self.alignSequenceA += self.sequenceA[idraw - 1]
                    self.alignSequenceB += "-"
                    idraw -= 1
                elif way == "L":
                    self.alignSequenceA += "-"
                    self.alignSequenceB += self.sequenceB[idcol - 1]
                    idcol -= 1
                else:
                    self.alignSequenceA += self.sequenceA[idraw - 1]
                    self.alignSequenceB += self.sequenceB[idcol - 1]
                    idraw -= 1
                    idcol -= 1
                self.findAlign(idraw, idcol)
                self.alignSequenceA = self.alignSequenceA[:-1]
                self.alignSequenceB = self.alignSequenceB[:-1]
                idraw, idcol = temp_idraw, temp_idcol


class LocalAlign(Align):
    def __init__(self, sequenceA, sequenceB, substitutionMatrixName="blosum62.txt", I=12, E=2, numberSeqToFind=1):

        Align.__init__(self, sequenceA, sequenceB, substitutionMatrixName, I, E,"True")
        self.fillMatrices()
        self.backtrackMatrixTrace = self.createBacktrackMatrixTrace()
        for k in range(numberSeqToFind):
            self.alignSequenceA = ""
            self.alignSequenceB = ""
            maxPos = self.findMaxPos()
            if maxPos == (0, 0):
                # Matrice S contient que des 0
                # -1 = valeur sentinelle pour dire plus possible de trouver des séquences.
                self.solutions.append(-1)
                break
            else:
                self.allMaximunPos.append(maxPos)
                visitedPos = list()
                self.findAlign(maxPos, visitedPos)
                self.allVisitedPos.update(visitedPos)
                self.recreateMatrix(visitedPos)
                self.backtrackMatrixTrace = self.createBacktrackMatrixTrace()


    def recreateMatrix(self, visitedPos):

        self.clearPath(visitedPos)
        self.fillMatrices()

    def clearPath(self, visitedPos):
        for tuplePos in visitedPos:
            self.S[tuplePos[0]][tuplePos[1]] = 0
            self.V[tuplePos[0]][tuplePos[1]] = 0
            self.W[tuplePos[0]][tuplePos[1]] = 0

    def findMaxPos(self):
        maximum = 0
        maxPos = (0, 0)
        for idraw in range(1, self.lineSize + 1):
            for idcol in range(1, self.columnSize + 1):
                if (idraw, idcol) not in self.allMaximunPos and (self.S[idraw][idcol] > maximum):
                    maxPos = (idraw, idcol)
                    maximum = self.S[idraw][idcol]
        return maxPos

    def findAlign(self, maxPos, visitedPos):
        idraw, idcol = maxPos[0], maxPos[1]
        while self.S[idraw][idcol] != 0:
            visitedPos.append((idraw, idcol))

            if self.backtrackMatrixTrace[idraw][idcol] == "T":
                self.alignSequenceA += self.sequenceA[idraw - 1]
                self.alignSequenceB += "-"
                idraw -= 1
            elif self.backtrackMatrixTrace[idraw][idcol] == "L":
                self.alignSequenceA += "-"
                self.alignSequenceB += self.sequenceB[idcol - 1]
                idcol -= 1
            else:
                self.alignSequenceA += self.sequenceA[idraw - 1]
                self.alignSequenceB += self.sequenceB[idcol - 1]
                idraw -= 1
                idcol -= 1

        self.solutions.append((self.alignSequenceA[::-1], self.alignSequenceB[::-1], self.S[maxPos[0]][maxPos[1]],
        visitedPos[-1][0]-1, visitedPos[0][0]-1, visitedPos[-1][1]-1, visitedPos[0][1]-1))


def getBestGlobalAlign():
    fileName = "WW-sequence.fasta"
    sequences = Sequence.load(fileName)
    allGlobalAlign = []
    for i in sequences:
        sequenceA = i.getSequence()
        for j in sequences[sequences.index(i)+1:]:
            sequenceB = j.getSequence()
            allGlobalAlign.append(GlobalAlign(sequenceA, sequenceB))

    allscore = [globalAlign.solutions[0][2] for globalAlign in allGlobalAlign]
    posMax = allscore.index(max(allscore))
    print(allGlobalAlign[posMax])


def getBestLocalAlign():
    fileName = "protein-sequences.fasta"
    sequences = Sequence.load(fileName)
    allLocalAlign = []
    for i in sequences:
        sequenceA = i.getSequence()
        for j in sequences[sequences.index(i)+1:]:
            sequenceB = j.getSequence()
            allLocalAlign.append(LocalAlign(sequenceA, sequenceB))
    allscore = [localAlign.solutions[0][2] for localAlign in allLocalAlign]
    posMax = allscore.index(max(allscore))
    print(allLocalAlign[posMax])


if __name__ == '__main__':
    #getBestGlobalAlign()
    #getBestLocalAlign()
    #sequences = Sequence.load("WW-sequence.fasta")
    testAlign = GlobalAlign("THISLINE", "ISALIGNED", substitutionMatrixName="blosum62.txt")
    #testAlign = GlobalAlign(sequences[0].getSequence(), sequences[4].getSequence(), substitutionMatrixName="blosum62.txt", numberSeqToFind=2)
    #testAlign = GlobalAlign("GGVTTT", "MGGETFA", substitutionMatrixName="blosum62.txt", I=12, E=2, numberSeqToFind = 2 )
    #sequences = Sequence.load("protein-sequences.fasta")
    # testAlign = LocalAlign("THISLINE", "ISALIGNED", I = 4, E = 1, substitutionMatrixName="blosum62.txt")
    # testAlign = LocalAlign("WRITERS","VINTNER","blosum62.txt", I=12, E=2, numberSeqToFind=10)
    # print(testAlign)
    #testAlign = LocalAlign(sequences[0].getSequence(), sequences[1].getSequence(), I=4, E=1,
    #                     substitutionMatrixName="blosum62.txt", numberSeqToFind=2)
    print(testAlign)