#!/usr/bin/python3
# -*- coding: utf-8 -*-


class Score():
    """
    Classe représentant un ADT Score (matrice de substitution) et opérations qu'on peut faire dessus.
    """
    lettersIndexes = ['A', 'R', 'N',
                      'D', 'C', 'Q',
                      'E', 'G', 'H',
                      'I', 'L', 'K',
                      'M', 'F', 'P',
                      'S', 'T', 'W',
                      'Y', 'V', 'B',
                      'Z', 'X', '*']


    def __init__(self, matrix):
        self.matrix = matrix


    def __repr__(self):
        """
        Fonction qui imprime joliment la matrice, avec le nom de chaque colonne/liste
        """
        stringMatrix = str()
        stringMatrix += (" " * 5)
        for letter in self.lettersIndexes:
            stringMatrix += "{:^5}".format(letter) #{:^5} est formatage pour avoir des colonnes/lignes identiques.
        stringMatrix += "\n"
        for idraw in range(len(self.matrix)):
            stringMatrix += "{:^5}".format(self.lettersIndexes[idraw])
            for item in self.matrix[idraw]:
                stringMatrix += "{:^5}".format(item)
            stringMatrix += "\n"
        return stringMatrix


    def __str__(self):
        return repr(self)


    def __getitem__(self, aminoAcid):
        """
        Fonction inspiré de __getitem__ de la classe list.
        Renvoie le score de substitution  de 2 acide aminé
        contenu dans un tuple aminoAcid.
        """
        idraw = self.lettersIndexes.index(aminoAcid[0])
        idcol = self.lettersIndexes.index(aminoAcid[1])
        return int(self.matrix[idraw][idcol])


    @classmethod
    def loadData(cls, fileName):
        """
        Parser qui charge la matrice de substitution depuis un fichier blosum ou pam
        """
        with open(fileName) as fd:
            matrix = []
            line = fd.readline()
            while line[0] == '#':
                line = fd.readline()
            line = fd.readline() #SKIP Ligne avec le nom des colonnes.
            while line.strip("\n")!="":
                line = line.split()
                # ne prends pas le 1er caratère nom de la ligne.
                matrix.append(line[1:])
                line = fd.readline()
        return Score(matrix)

if __name__ == '__main__':
    test_score = Score.loadData("pam120.txt")
    print(test_score)
