#!/usr/bin/python3
# -*- coding: utf-8 -*-

class Sequence:
    """
    ADT Sequence qui représente une séquence d'acides aminés et toutes les opérations qu'on peut faire sur une séquence.
    """
    def __init__(self, sequence):
        self.sequence = sequence

    def __getitem__(self):
        return self.sequence

    def getSequence(self):
        """
        Retourne le string de la séquence.
        """
        return self.sequence

    def __len__(self):
        return len(self.sequence)

    def __repr__(self):
        """
        Représente ce que contient la classe.
        """
        sequences = "{}\n".format(self.sequence)
        return sequences


    @classmethod
    def load(cls, fileName):
        """
        Parser qui charge chaque séquence d'acide aminé
        dans une instance de Sequence, mis dans une liste.
        fileName: paramètre qui est le fichier fasta contenant les séquences à traiter
        """
        with open(fileName) as fd:
            listSequences = []
            line = fd.readline()
            while line:
                sequence = ""
                if line[0] == ">":
                    line = fd.readline() #SKIP la ligne d'identification
                    while line and line[0] != ">" :
                        #Boucle si séquence disposé sur plusieurs lignes
                        sequence += line.strip("\n")
                        line = fd.readline()
                    if sequence: #Si séquence trouvé, on ajoute.
                        listSequences.append(cls(sequence.strip("\n")))
        return listSequences

if __name__ == '__main__':
    test_sequence = Sequence.load("protein-sequences.fasta")
    print(test_sequence)
