#!/usr/bin/python3

class Sequence:
    """
    ADT Sequence qui représente une séquence d'acides aminés et toutes les opérations qu'on peut faire sur une séquence.
    """

    def __init__(self, identificator, sequence, structure):
        self.sequence = (identificator, sequence, structure)

    def getSequence(self):
        """
        Retourne le string de la séquence.
        """
        return self.sequence[1]

    def getIdentificator(self):
        """
        Retourne l'identifiant de la séquence.
        """
        return self.sequence[0]

    def getStructure(self):
        """
        Retourne la structure secondaire de la séquence.
        """
        return self.sequence[2]

    def __len__(self):
        """
        Retourne la taille d'une séquence d'acides aminés.
        """
        return len(self.sequence[1])

    def __repr__(self):
        """
        Représente ce que contient la classe.
        """
        sequences = "Identifiant: {} \n Séquence: {}\n Structure Secondaire {}\n".format(self.sequence[0], self.sequence[1], self.sequence[2])
        return sequences

    @classmethod
    def loadData(cls, fileName):
        """
        Parser qui charge chaque séquence d'acide aminé
        dans une instance de Sequence, mis dans une liste.
        fileName: paramètre qui est le fichier fasta contenant les séquences à traiter
        """
        with open(fileName) as fd:
            listSequences = []
            line = fd.readline()
            while line:
                sequence = ""
                structure = ""
                if line[0] == ">":
                    identificator = line[1:].strip("\n")
                    line = fd.readline()  # SKIP la ligne d'identification
                    while line and line[0] != ">":
                        # Boucle si séquence disposé sur plusieurs lignes
                        sequence += line.strip("\n")
                        line = fd.readline()
                        structure += line.strip("\n")
                        line = fd.readline()
                    if sequence:  # Si séquence trouvé, on ajoute.
                        listSequences.append(cls(identificator, sequence.strip("\n"), structure))

        return listSequences

if __name__ == '__main__':
    test = Sequence.loadData("dataFromDssp_test.fasta")
    print(test)